from django.test import TestCase

from .forms import AuthorForm
from .models import Address, Author


class ModelFormTests(TestCase):
    def test_update(self):
        author = Author.objects.create(
            name="Bob", age=50, address=Address(city="NYC", state="NY", zip_code="10001")
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "New York City",
            "address-state": "NY",
            "address-zip_code": "10001",
        }
        form = AuthorForm(data, instance=author)
        self.assertTrue(form.is_valid())
        form.save()
        author.refresh_from_db()
        self.assertEqual(author.age, 51)
        self.assertEqual(author.address.city, "New York City")

    def test_some_missing_data(self):
        author = Author.objects.create(
            name="Bob", age=50, address=Address(city="NYC", state="NY", zip_code="10001")
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "New York City",
            "address-state": "NY",
            "address-zip_code": "",
        }
        form = AuthorForm(data, instance=author)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["address"], ["Enter all required values."])

    def test_invalid_field_data(self):
        """A field's data (state) is too long."""
        author = Author.objects.create(
            name="Bob", age=50, address=Address(city="NYC", state="NY", zip_code="10001")
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "New York City",
            "address-state": "TOO LONG",
            "address-zip_code": "",
        }
        form = AuthorForm(data, instance=author)
        self.assertFalse(form.is_valid())
        self.assertEqual(
            form.errors["address"],
            [
                "Ensure this value has at most 2 characters (it has 8).",
                "Enter all required values.",
            ],
        )

    def test_all_missing_data(self):
        author = Author.objects.create(
            name="Bob", age=50, address=Address(city="NYC", state="NY", zip_code="10001")
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "",
            "address-state": "",
            "address-zip_code": "",
        }
        form = AuthorForm(data, instance=author)
        self.assertFalse(form.is_valid())
        self.assertEqual(form.errors["address"], ["This field is required."])

    def test_nullable_field(self):
        """A nullable EmbeddedModelField is removed if all fields are empty."""
        author = Author.objects.create(
            name="Bob",
            age=50,
            address=Address(city="NYC", state="NY", zip_code="10001"),
            billing_address=Address(city="NYC", state="NY", zip_code="10001"),
        )
        data = {
            "name": "Bob",
            "age": 51,
            "address-po_box": "",
            "address-city": "New York City",
            "address-state": "NY",
            "address-zip_code": "10001",
            "billing_address-po_box": "",
            "billing_address-city": "",
            "billing_address-state": "",
            "billing_address-zip_code": "",
        }
        form = AuthorForm(data, instance=author)
        self.assertTrue(form.is_valid())
        form.save()
        author.refresh_from_db()
        self.assertIsNone(author.billing_address)

    def test_rendering(self):
        form = AuthorForm()
        self.assertHTMLEqual(
            str(form.fields["address"].get_bound_field(form, "address")),
            """
            <div>
                <label for="id_address-po_box">PO Box:</label>
                <input id="id_address-po_box" maxlength="50" name="address-po_box" type="text">
            </div>
            <div>
                <label for="id_address-city">City:</label>
                <input type="text" name="address-city" maxlength="20" required id="id_address-city">
            </div>
            <div>
                <label for="id_address-state">State:</label>
                <input type="text" name="address-state" maxlength="2" required
                    id="id_address-state">
            </div>
            <div>
                <label for="id_address-zip_code">Zip code:</label>
                <input type="number" name="address-zip_code" required id="id_address-zip_code">
            </div>""",
        )



from django.db import NotSupportedError
from django.db.models import Index
from django.db.models.fields.related_lookups import In
from django.db.models.lookups import BuiltinLookup
from django.db.models.sql.query import Query
from django.db.models.sql.where import AND, XOR, WhereNode

from .query_utils import process_rhs

MONGO_INDEX_OPERATORS = {
    "exact": "$eq",
    "gt": "$gt",
    "gte": "$gte",
    "lt": "$lt",
    "lte": "$lte",
    "in": "$in",
}


def _get_condition_mql(self, model, schema_editor):
    """Analogous to Index._get_condition_sql()."""
    query = Query(model=model, alias_cols=False)
    where = query.build_where(self.condition)
    compiler = query.get_compiler(connection=schema_editor.connection)
    return where.as_mql_idx(compiler, schema_editor.connection)


def builtin_lookup_idx(self, compiler, connection):
    lhs_mql = self.lhs.target.column
    value = process_rhs(self, compiler, connection)
    try:
        operator = MONGO_INDEX_OPERATORS[self.lookup_name]
    except KeyError:
        raise NotSupportedError(
            f"MongoDB does not support the '{self.lookup_name}' lookup in indexes."
        ) from None
    return {lhs_mql: {operator: value}}


def in_idx(self, compiler, connection):
    if not connection.features.supports_in_index_operator:
        raise NotSupportedError("MongoDB < 6.0 does not support the 'in' lookup in indexes.")
    return builtin_lookup_idx(self, compiler, connection)


def where_node_idx(self, compiler, connection):
    if self.connector == AND:
        operator = "$and"
    elif self.connector == XOR:
        raise NotSupportedError("MongoDB does not support the '^' operator lookup in indexes.")
    else:
        if not connection.features.supports_in_index_operator:
            raise NotSupportedError("MongoDB < 6.0 does not support the '|' operator in indexes.")
        operator = "$or"
    if self.negated:
        raise NotSupportedError("MongoDB does not support the '~' operator in indexes.")
    children_mql = []
    for child in self.children:
        mql = child.as_mql_idx(compiler, connection)
        children_mql.append(mql)
    if len(children_mql) == 1:
        mql = children_mql[0]
    elif len(children_mql) > 1:
        mql = {operator: children_mql}
    else:
        mql = {}
    return mql


def register_indexes():
    BuiltinLookup.as_mql_idx = builtin_lookup_idx
    In.as_mql_idx = in_idx
    Index._get_condition_mql = _get_condition_mql
    WhereNode.as_mql_idx = where_node_idx



DATABASES = {
    "default": {
        "ENGINE": "django_mongodb_backend",
        "NAME": "djangotests",
    },
    "other": {
        "ENGINE": "django_mongodb_backend",
        "NAME": "djangotests-other",
    },
}
DEFAULT_AUTO_FIELD = "django_mongodb_backend.fields.ObjectIdAutoField"
PASSWORD_HASHERS = ("django.contrib.auth.hashers.MD5PasswordHasher",)
SECRET_KEY = "django_tests_secret_key"
USE_TZ = False



from django.db import NotSupportedError
from django.db.models.expressions import Func
from django.db.models.functions.comparison import Cast, Coalesce, Greatest, Least, NullIf
from django.db.models.functions.datetime import (
    Extract,
    ExtractDay,
    ExtractHour,
    ExtractIsoWeekDay,
    ExtractIsoYear,
    ExtractMinute,
    ExtractMonth,
    ExtractSecond,
    ExtractWeek,
    ExtractWeekDay,
    ExtractYear,
    Now,
    TruncBase,
    TruncDate,
    TruncTime,
)
from django.db.models.functions.math import Ceil, Cot, Degrees, Log, Power, Radians, Random, Round
from django.db.models.functions.text import (
    Concat,
    ConcatPair,
    Left,
    Length,
    Lower,
    LTrim,
    Replace,
    RTrim,
    StrIndex,
    Substr,
    Trim,
    Upper,
)

from .query_utils import process_lhs

MONGO_OPERATORS = {
    Ceil: "ceil",
    Coalesce: "ifNull",
    Degrees: "radiansToDegrees",
    Greatest: "max",
    Least: "min",
    Power: "pow",
    Radians: "degreesToRadians",
    Random: "rand",
}
EXTRACT_OPERATORS = {
    ExtractDay.lookup_name: "dayOfMonth",
    ExtractHour.lookup_name: "hour",
    ExtractIsoWeekDay.lookup_name: "isoDayOfWeek",
    ExtractIsoYear.lookup_name: "isoWeekYear",
    ExtractMinute.lookup_name: "minute",
    ExtractMonth.lookup_name: "month",
    ExtractSecond.lookup_name: "second",
    ExtractWeek.lookup_name: "isoWeek",
    ExtractWeekDay.lookup_name: "dayOfWeek",
    ExtractYear.lookup_name: "year",
}


def cast(self, compiler, connection):
    output_type = connection.data_types[self.output_field.get_internal_type()]
    lhs_mql = process_lhs(self, compiler, connection)[0]
    if max_length := self.output_field.max_length:
        lhs_mql = {"$substrCP": [lhs_mql, 0, max_length]}
    # Skip the conversion for "object" as it doesn't need to be transformed for
    # interpretation by JSONField, which can handle types including int,
    # object, or array.
    if output_type != "object":
        lhs_mql = {"$convert": {"input": lhs_mql, "to": output_type}}
    if decimal_places := getattr(self.output_field, "decimal_places", None):
        lhs_mql = {"$trunc": [lhs_mql, decimal_places]}
    return lhs_mql


def concat(self, compiler, connection):
    return self.get_source_expressions()[0].as_mql(compiler, connection)


def concat_pair(self, compiler, connection):
    # null on either side results in null for expression, wrap with coalesce.
    coalesced = self.coalesce()
    return super(ConcatPair, coalesced).as_mql(compiler, connection)


def cot(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    return {"$divide": [1, {"$tan": lhs_mql}]}


def extract(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    operator = EXTRACT_OPERATORS.get(self.lookup_name)
    if operator is None:
        raise NotSupportedError("%s is not supported." % self.__class__.__name__)
    if timezone := self.get_tzname():
        lhs_mql = {"date": lhs_mql, "timezone": timezone}
    return {f"${operator}": lhs_mql}


def func(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    operator = MONGO_OPERATORS.get(self.__class__, self.function.lower())
    return {f"${operator}": lhs_mql}


def left(self, compiler, connection):
    return self.get_substr().as_mql(compiler, connection)


def length(self, compiler, connection):
    # Check for null first since $strLenCP only accepts strings.
    lhs_mql = process_lhs(self, compiler, connection)
    return {"$cond": {"if": {"$eq": [lhs_mql, None]}, "then": None, "else": {"$strLenCP": lhs_mql}}}


def log(self, compiler, connection):
    # This function is usually log(base, num) but on MongoDB it's log(num, base).
    clone = self.copy()
    clone.set_source_expressions(self.get_source_expressions()[::-1])
    return func(clone, compiler, connection)


def now(self, compiler, connection):  # noqa: ARG001
    return "$$NOW"


def null_if(self, compiler, connection):
    """Return None if expr1==expr2 else expr1."""
    expr1, expr2 = (expr.as_mql(compiler, connection) for expr in self.get_source_expressions())
    return {"$cond": {"if": {"$eq": [expr1, expr2]}, "then": None, "else": expr1}}


def preserve_null(operator):
    # If the argument is null, the function should return null, not
    # $toLower/Upper's behavior of returning an empty string.
    def wrapped(self, compiler, connection):
        lhs_mql = process_lhs(self, compiler, connection)
        return {
            "$cond": {
                "if": connection.mongo_operators["isnull"](lhs_mql, True),
                "then": None,
                "else": {f"${operator}": lhs_mql},
            }
        }

    return wrapped


def replace(self, compiler, connection):
    expression, text, replacement = process_lhs(self, compiler, connection)
    return {"$replaceAll": {"input": expression, "find": text, "replacement": replacement}}


def round_(self, compiler, connection):
    # Round needs its own function because it's a special case that inherits
    # from Transform but has two arguments.
    return {"$round": [expr.as_mql(compiler, connection) for expr in self.get_source_expressions()]}


def str_index(self, compiler, connection):
    lhs = process_lhs(self, compiler, connection)
    # StrIndex should be 0-indexed (not found) but it's -1-indexed on MongoDB.
    return {"$add": [{"$indexOfCP": lhs}, 1]}


def substr(self, compiler, connection):
    lhs = process_lhs(self, compiler, connection)
    # The starting index is zero-indexed on MongoDB rather than one-indexed.
    lhs[1] = {"$add": [lhs[1], -1]}
    # If no limit is specified, use the length of the string since $substrCP
    # requires one.
    if len(lhs) == 2:
        lhs.append({"$strLenCP": lhs[0]})
    return {"$substrCP": lhs}


def trim(operator):
    def wrapped(self, compiler, connection):
        lhs = process_lhs(self, compiler, connection)
        return {f"${operator}": {"input": lhs}}

    return wrapped


def trunc(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    lhs_mql = {"date": lhs_mql, "unit": self.kind, "startOfWeek": "mon"}
    if timezone := self.get_tzname():
        lhs_mql["timezone"] = timezone
    return {"$dateTrunc": lhs_mql}


def trunc_date(self, compiler, connection):
    # Cast to date rather than truncate to date.
    lhs_mql = process_lhs(self, compiler, connection)
    tzname = self.get_tzname()
    if tzname and tzname != "UTC":
        raise NotSupportedError(f"TruncDate with tzinfo ({tzname}) isn't supported on MongoDB.")
    return {
        "$dateFromString": {
            "dateString": {
                "$concat": [
                    {"$dateToString": {"format": "%Y-%m-%d", "date": lhs_mql}},
                    # Dates are stored with time(0, 0), so by replacing any
                    # existing time component with that, the result of
                    # TruncDate can be compared to DateField.
                    "T00:00:00.000",
                ]
            },
        }
    }


def trunc_time(self, compiler, connection):
    lhs_mql = process_lhs(self, compiler, connection)
    return {
        "$dateFromString": {
            "dateString": {
                "$concat": [
                    # Times are stored with date(1, 1, 1)), so by
                    # replacing any existing date component with that, the
                    # result of TruncTime can be compared to TimeField.
                    "0001-01-01T",
                    {"$dateToString": {"format": "%H:%M:%S.%L", "date": lhs_mql}},
                ]
            }
        }
    }


def register_functions():
    Cast.as_mql = cast
    Concat.as_mql = concat
    ConcatPair.as_mql = concat_pair
    Cot.as_mql = cot
    Extract.as_mql = extract
    Func.as_mql = func
    Left.as_mql = left
    Length.as_mql = length
    Log.as_mql = log
    Lower.as_mql = preserve_null("toLower")
    LTrim.as_mql = trim("ltrim")
    Now.as_mql = now
    NullIf.as_mql = null_if
    Replace.as_mql = replace
    Round.as_mql = round_
    RTrim.as_mql = trim("rtrim")
    StrIndex.as_mql = str_index
    Substr.as_mql = substr
    Trim.as_mql = trim("trim")
    TruncBase.as_mql = trunc
    TruncDate.as_mql = trunc_date
    TruncTime.as_mql = trunc_time
    Upper.as_mql = preserve_null("toUpper")



from bson import ObjectId, errors
from django.core import exceptions
from django.db.models.fields import AutoField
from django.utils.functional import cached_property

from .objectid import ObjectIdMixin


class ObjectIdAutoField(ObjectIdMixin, AutoField):
    def __init__(self, *args, **kwargs):
        kwargs["db_column"] = "_id"
        super().__init__(*args, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if self.db_column == "_id":
            del kwargs["db_column"]
        if path.startswith("django_mongodb_backend.fields.auto"):
            path = path.replace(
                "django_mongodb_backend.fields.auto", "django_mongodb_backend.fields"
            )
        return name, path, args, kwargs

    def get_prep_value(self, value):
        if value is None:
            return None
        # Accept int for compatibility with Django's test suite which has many
        # instances of manually assigned integer IDs, as well as for things
        # like settings.SITE_ID which has a system check requiring an integer.
        if isinstance(value, (ObjectId | int)):
            return value
        try:
            return ObjectId(value)
        except errors.InvalidId as e:
            # A manually assigned integer ID?
            if isinstance(value, str) and value.isdigit():
                return int(value)
            raise ValueError(f"Field '{self.name}' expected an ObjectId but got {value!r}.") from e

    def get_internal_type(self):
        return "ObjectIdAutoField"

    def to_python(self, value):
        if value is None or isinstance(value, int):
            return value
        try:
            return ObjectId(value)
        except errors.InvalidId:
            try:
                return int(value)
            except ValueError:
                raise exceptions.ValidationError(
                    self.error_messages["invalid"],
                    code="invalid",
                    params={"value": value},
                ) from None

    @cached_property
    def validators(self):
        # Avoid IntegerField validators inherited from AutoField.
        return [*self.default_validators, *self._validators]



from django.core.exceptions import ValidationError
from django.db import models
from django.test import SimpleTestCase, TestCase
from django.test.utils import isolate_apps

from django_mongodb_backend.fields import EmbeddedModelField

from .models import (
    Address,
    Author,
    Book,
    Data,
    Holder,
)


class MethodTests(SimpleTestCase):
    def test_deconstruct(self):
        field = EmbeddedModelField("Data", null=True)
        name, path, args, kwargs = field.deconstruct()
        self.assertEqual(path, "django_mongodb_backend.fields.EmbeddedModelField")
        self.assertEqual(args, [])
        self.assertEqual(kwargs, {"embedded_model": "Data", "null": True})

    def test_get_db_prep_save_invalid(self):
        msg = "Expected instance of type <class 'model_fields_.models.Data'>, " "not <class 'int'>."
        with self.assertRaisesMessage(TypeError, msg):
            Holder(data=42).save()

    def test_validate(self):
        obj = Holder(data=Data(integer=None))
        # This isn't quite right because "integer" is the subfield of data
        # that's non-null.
        msg = "{'data': ['This field cannot be null.']}"
        with self.assertRaisesMessage(ValidationError, msg):
            obj.full_clean()


class ModelTests(TestCase):
    def truncate_ms(self, value):
        """Truncate microseconds to milliseconds as supported by MongoDB."""
        return value.replace(microsecond=(value.microsecond // 1000) * 1000)

    def test_save_load(self):
        Holder.objects.create(data=Data(integer="5"))
        obj = Holder.objects.get()
        self.assertIsInstance(obj.data, Data)
        # get_prep_value() is called, transforming string to int.
        self.assertEqual(obj.data.integer, 5)
        # Primary keys should not be populated...
        self.assertEqual(obj.data.id, None)
        # ... unless set explicitly.
        obj.data.id = obj.id
        obj.save()
        obj = Holder.objects.get()
        self.assertEqual(obj.data.id, obj.id)

    def test_save_load_null(self):
        Holder.objects.create(data=None)
        obj = Holder.objects.get()
        self.assertIsNone(obj.data)

    def test_pre_save(self):
        """Field.pre_save() is called on embedded model fields."""
        obj = Holder.objects.create(data=Data())
        auto_now = self.truncate_ms(obj.data.auto_now)
        auto_now_add = self.truncate_ms(obj.data.auto_now_add)
        self.assertEqual(auto_now, auto_now_add)
        # save() updates auto_now but not auto_now_add.
        obj.save()
        self.assertEqual(self.truncate_ms(obj.data.auto_now_add), auto_now_add)
        auto_now_two = obj.data.auto_now
        self.assertGreater(auto_now_two, obj.data.auto_now_add)
        # And again, save() updates auto_now but not auto_now_add.
        obj = Holder.objects.get()
        obj.save()
        self.assertEqual(obj.data.auto_now_add, auto_now_add)
        self.assertGreater(obj.data.auto_now, auto_now_two)


class QueryingTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.objs = [Holder.objects.create(data=Data(integer=x)) for x in range(6)]

    def test_exact(self):
        self.assertCountEqual(Holder.objects.filter(data__integer=3), [self.objs[3]])

    def test_lt(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__lt=3), self.objs[:3])

    def test_lte(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__lte=3), self.objs[:4])

    def test_gt(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__gt=3), self.objs[4:])

    def test_gte(self):
        self.assertCountEqual(Holder.objects.filter(data__integer__gte=3), self.objs[3:])

    def test_nested(self):
        obj = Book.objects.create(
            author=Author(name="Shakespeare", age=55, address=Address(city="NYC", state="NY"))
        )
        self.assertCountEqual(Book.objects.filter(author__address__city="NYC"), [obj])


@isolate_apps("model_fields_")
class CheckTests(SimpleTestCase):
    def test_no_relational_fields(self):
        class Target(models.Model):
            key = models.ForeignKey("MyModel", models.CASCADE)

        class MyModel(models.Model):
            field = EmbeddedModelField(Target)

        model = MyModel()
        errors = model.check()
        self.assertEqual(len(errors), 1)
        # The inner CharField has a non-positive max_length.
        self.assertEqual(errors[0].id, "django_mongodb.embedded_model.E001")
        msg = errors[0].msg
        self.assertEqual(
            msg, "Embedded models cannot have relational fields (Target.key is a ForeignKey)."
        )



from django.core.exceptions import ImproperlyConfigured
from django.db import connection
from django.test import SimpleTestCase

from django_mongodb_backend.base import DatabaseWrapper


class DatabaseWrapperTests(SimpleTestCase):
    def test_database_name_empty(self):
        settings = connection.settings_dict.copy()
        settings["NAME"] = ""
        msg = 'settings.DATABASES is missing the "NAME" value.'
        with self.assertRaisesMessage(ImproperlyConfigured, msg):
            DatabaseWrapper(settings).get_connection_params()



from django.conf import settings
from django.db.backends.base.creation import BaseDatabaseCreation


class DatabaseCreation(BaseDatabaseCreation):
    def _execute_create_test_db(self, cursor, parameters, keepdb=False):
        if not keepdb:
            self._destroy_test_db(parameters["dbname"], verbosity=0)

    def _destroy_test_db(self, test_database_name, verbosity):
        # At this point, settings still points to the non-test database. For
        # MongoDB, it must use the test database.
        settings.DATABASES[self.connection.alias]["NAME"] = test_database_name
        self.connection.settings_dict["NAME"] = test_database_name

        for collection in self.connection.introspection.table_names():
            if not collection.startswith("system."):
                self.connection.database.drop_collection(collection)
